-- read audio cd
-- v1.6 Feb 26, 2017

-- audio cd sector size = 2352 bytes
-- 2 x BRAM of 16384 bytes to hold 6 sectors each = 14112 bytes
-- 1 BRAM is filled with data from cd while the other is dma'ed to ddr
-- should not exceed 65536 -> 4 times for a read_cd command (=56448 bytes) 
-- app needs chunks of 64K 16-bit stereo samples = 262144 bytes
-- so read 5 x 4 x 6 x 2352 = 282240 bytes for a page

-- sector = 2352 bytes
-- block = 6 sectors (14112 bytes)
-- chunk = 4 blocks (56448 bytes)
-- page = 5 chunks (282240 bytes)

-- clk = 80 MHz (12.5 ns)
-- fast read = 7 + 4 = 11 cycles (137.5 ns)
-- adjusted with delay_count

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_unsigned.all;

entity atapi is
  Port(
    clk : in std_logic;
    data : inout std_logic_vector(15 downto 0);
    dmarq : in std_logic;
    intrq : in std_logic;
    adr : out std_logic_vector(4 downto 0);
    diow : out std_logic;
    dior : out std_logic;
    dmack : out std_logic;

    start : in std_logic;
    err_code : out std_logic_vector(2 downto 0);
    track_num : in std_logic_vector(23 downto 0);
    len : in std_logic_vector(15 downto 0);
    
    mute : in std_logic;
    flip : out std_logic_vector(7 downto 0);
    BRAM0_addr : out std_logic_vector(31 downto 0);
    BRAM0_data : out std_logic_vector(31 downto 0);
    BRAM0_we : out std_logic_vector(3 downto 0);
    BRAM1_addr : out std_logic_vector(31 downto 0);
    BRAM1_data : out std_logic_vector(31 downto 0);
    BRAM1_we : out std_logic_vector(3 downto 0)
  );
end atapi;

architecture Behavioral of atapi is
  
  type send_packet_cmd_states is (IDLE, READ_CD_CMD, SEND_CMD, ALT_REG, WAIT_BUSY_C_1, WAIT_DRQ_S, GET_REASON_1, SET_SIZE_L, SET_SIZE_H, SEND_PACKET, WAIT_BUSY_C_2, POLL_DRQ_C_2, POLL_DRQ_C_W2, GET_REASON_2, GET_REASON_3, GET_SIZE_L, GET_SIZE_H, GET_SIZE, GET_RESP, GET_RESP_1, DONE, ERR, READ, WRITE, NEXT_STAT );
  signal exec_state : send_packet_cmd_states := IDLE;  
  signal next_state : send_packet_cmd_states := IDLE; 

  type array_packet is array (0 to 11) of std_logic_vector(7 downto 0); 
  signal array_pack : array_packet;

  signal bram_addr : std_logic_vector(19 downto 0) := (others => '0');
  signal we0 : std_logic_vector(3 downto 0) := (others => '0');
  signal we1 : std_logic_vector(3 downto 0) := (others => '0');
  signal adrsel : std_logic := '0';
  signal size : std_logic_vector(15 downto 0) := (others => '0'); 
  signal packet_num : std_logic_vector(2 downto 0) := (others => '0'); 

  -- IDE interface
  signal data_out : std_logic_vector(15 downto 0) := (others => '0');
  signal data_in : std_logic_vector(15 downto 0);
  signal data_dir : std_logic := '1';
  signal ata_reg : std_logic_vector(4 downto 0);
  signal diow_s : std_logic;
  signal dior_s : std_logic;
  
  signal err_code_s : std_logic_vector(2 downto 0) := (others => '0');
  signal delay_count : std_logic_vector(4 downto 0) := (others => '0');
  signal timeout_valid : std_logic := '0';
  signal last_timeout_valid : std_logic := '0';
  signal timeout : std_logic_vector(21 downto 0) := (others => '0');
  signal flip_s : std_logic := '0';
  signal block_count : std_logic_vector(2 downto 0) := (others => '0');
  signal chunk_count : std_logic_vector(3 downto 0) := (others => '0'); 
  signal fast_read : std_logic := '0';
  signal track : std_logic_vector(23 downto 0) := (others => '0'); 
  signal sample : std_logic_vector(15 downto 0) := (others => '0'); 
 
begin

BRAM0_addr <= x"0000" & "00" & bram_addr(13 downto 0);  -- 4096*4=16384
BRAM1_addr <= x"0000" & "00" & bram_addr(13 downto 0);
BRAM0_we <= we0;
BRAM1_we <= we1; 

data <= data_out when data_dir = '0' else (others=>'Z');
adr <= ata_reg;
diow <= diow_s;
dior <= dior_s;
dmack <= '1';  -- unused for now
flip <= "01" & "00000" & flip_s; -- re-use from_spdif ID = 1;
err_code <= err_code_s;

process(clk)
begin
  if clk'event and clk = '1' then
    data_dir <= '1';  -- all inputs
    diow_s <= '1';  -- negated
    dior_s <= '1';  -- negated
    we0 <= "0000";
    we1 <= "0000";

    if (timeout_valid = '1') and (last_timeout_valid = '0') then
      timeout <= (others => '1');  -- start counter
    end if;
    last_timeout_valid <= timeout_valid;
    
    case exec_state is 
    when IDLE =>
      timeout_valid <= '0';
      if start = '1' then
        block_count <= "000"; -- 6 blocks of 2352 bytes for a bram
        timeout_valid <= '0';
        array_pack(0) <= x"be";  -- READ_CD command
        array_pack(1) <= x"00";
        array_pack(2) <= x"00";
        array_pack(3) <= x"00";
        array_pack(4) <= x"00";
        array_pack(5) <= x"00";
        array_pack(6) <= x"00";
        array_pack(7) <= x"00";
        array_pack(8) <= x"18"; -- 24 (4 x 6 x 2352)
        array_pack(9) <= x"f0";  -- user data
        array_pack(10) <= x"00";
        array_pack(11) <= x"00"; 
        chunk_count <= "0101";  -- 5 times (x 4 x 6 x 2352)
        track <= track_num;
        exec_state <= READ_CD_CMD;
      end if;
        
    when READ_CD_CMD =>
      if chunk_count = "0000" then
        exec_state <= DONE;
      else
        chunk_count <= chunk_count - 1;
        array_pack(3) <= track(23 downto 16);
        array_pack(4) <= track(15 downto 8);
        array_pack(5) <= track(7 downto 0); 
        exec_state <= SEND_CMD;
      end if;
      
    -- write CMD_PACKET  ata command
    when SEND_CMD =>  
      track <= track + x"18";  -- update track number  
      err_code_s <= "111";  -- in progress
      ata_reg <= "10111";  -- cmd_stat_reg
      data_out <= x"ffa0";  -- command cmd_packet    
      exec_state <= WRITE;
      next_state <= ALT_REG;

    when ALT_REG =>
      ata_reg <= "01111";  -- alt_reg
      exec_state <= READ;
      timeout_valid <= '1';  -- arm counter
      next_state <= WAIT_BUSY_C_1;
    
    -- wait for BUSY to clear
    when WAIT_BUSY_C_1 =>
      ata_reg <= "10111";  -- cmd_stat_reg
      exec_state <= READ;
      next_state <= WAIT_BUSY_C_1; -- loop
      if (data_in and x"0080") = x"0000" then
        next_state <= WAIT_DRQ_S;
      end if;

    -- wait for DRQ to get set
    when WAIT_DRQ_S =>
      ata_reg <= "10111";  -- cmd_stat_reg
      exec_state <= READ;
      next_state <= WAIT_DRQ_S;  -- loop
      if (data_in and x"0008") /= x"0000" then
        ata_reg <= "10010";  -- sec_cnt_reg
        timeout_valid <= '0';  -- disarm counter
        next_state <= GET_REASON_1;
      end if;
      
    -- check interrupt reason reg for ATAPI_READY_TO_ACCEPT_PACKET
    when GET_REASON_1 =>
      exec_state <= READ;
      next_state <= SET_SIZE_L;
      if (data_in and x"0007") /= x"0001" then -- ATAPI_READY_TO_ACCEPT_PACKET
        err_code_s <= "001";
        next_state <= ERR;
      end if;

    -- default to max size LSB
    when SET_SIZE_L =>
      data_out <= x"ffff";  -- max size
      ata_reg <= "10100";  -- cyl_low_reg
      exec_state <= WRITE;
      next_state <= SET_SIZE_H;

    -- default to max size MSB
    when SET_SIZE_H =>
      data_out <= x"ffff";  -- max size
      ata_reg <= "10101";  -- cyl_high_reg
      packet_num <= "000";  -- 6 words in a packet
      exec_state <= WRITE;
      next_state <= SEND_PACKET;

    -- write 6 words packet
    when SEND_PACKET =>
      if packet_num < 6 then
        case packet_num is 
        when "000" => data_out <= array_pack(1) & array_pack(0);
        when "001" => data_out <= array_pack(3) & array_pack(2);
        when "010" => data_out <= array_pack(5) & array_pack(4);
        when "011" => data_out <= array_pack(7) & array_pack(6);
        when "100" => data_out <= array_pack(9) & array_pack(8);
        when "101" => data_out <= array_pack(11) & array_pack(10);
        when others =>
        end case;
        ata_reg <= "10000";  -- data_reg
        packet_num <= packet_num + 1;
        exec_state <= WRITE;
        next_state <= SEND_PACKET;
      else
        ata_reg <= "01111";  -- alt_reg
        timeout_valid <= '1';  -- arm counter
        exec_state <= READ;
        next_state <= WAIT_BUSY_C_2;
      end if;

    -- wait for BUSY to clear
    when WAIT_BUSY_C_2 =>
      ata_reg <= "10111";  -- cmd_stat_reg
      exec_state <= READ;
      next_state <= WAIT_BUSY_C_2;  -- loop
      if (data_in and x"0080") = x"0000" then
        exec_state <= POLL_DRQ_C_2;
      end if;
    
    -- check DRQ
    when POLL_DRQ_C_2 =>
      ata_reg <= "10111";  -- cmd_stat_reg
      exec_state <= READ;
      next_state <= POLL_DRQ_C_W2;
      if (data_in and x"0001") /= x"0000" then  -- check error bit
        err_code_s <= "010";
        exec_state <= ERR;
      end if;
      
    when POLL_DRQ_C_W2 =>
      ata_reg <= "10010";  -- sec_cnt_reg
      exec_state <= READ;
      if (data_in and x"0008") = x"0000" then  -- check DRQ
        next_state <= GET_REASON_2;
      else
        timeout_valid <= '0';  -- disarm counter
        next_state <= GET_REASON_3;
      end if;
    
    -- check interrupt reason reg for ATAPI_COMPLETION
    when GET_REASON_2 =>
      if (data_in and x"0007") = x"0003" then  -- ATAPI_COMPLETION
        exec_state <= READ_CD_CMD;
      else
        ata_reg <= "10111";  -- cmd_stat_reg
        exec_state <= READ;
        next_state <= WAIT_BUSY_C_2;
      end if;

    -- check interrupt reason reg for ATAPI_GET_DATA
    when GET_REASON_3 =>
      if (data_in and x"0007") /= x"0002" then  -- ATAPI_GET_DATA
        err_code_s <= "011";
        exec_state <= ERR;
      else
        ata_reg <= "10100";  -- cyl_low_reg
        exec_state <= READ;
        next_state <= GET_SIZE_L;
      end if;
      
    -- get data size LSB
    when GET_SIZE_L =>
      size(7 downto 0) <= data_in(7 downto 0);
      ata_reg <= "10101";  -- cyl_high_reg
      exec_state <= READ;
      next_state <= GET_SIZE_H;

    -- get data size MSB
    when GET_SIZE_H =>
      size(15 downto 8) <= data_in(7 downto 0);
      exec_state <= GET_SIZE;
      
    when GET_SIZE =>
      size <= '0' & size(15 downto 1);  -- divide by 2 for 16-bit samples
      ata_reg <= "10000";  -- data_reg
      fast_read <= '1';  -- engage the turbo!
      delay_count <= "00111";
      exec_state <= READ;
      next_state <= GET_RESP;

    -- read a sector of audio data
    when GET_RESP =>
      if adrsel = '0' then 
        if flip_s = '0' then 
          BRAM0_data <= x"0000" & sample;
          we0 <= "0011";
        else
          BRAM1_data <= x"0000" & sample;
          we1 <= "0011";
        end if;
      else
        if flip_s = '0' then 
          BRAM0_data <=  sample & x"0000";
          we0 <= "1100";
        else
          BRAM1_data <=  sample & x"0000";
          we1 <= "1100";
        end if;
      end if; 
      exec_state <= GET_RESP_1;
    
    -- update on following cycle  
    when GET_RESP_1 =>  
      if size = x"0000" then
        if block_count = "101" then
          block_count <= "000";
          bram_addr <= (others => '0');
          adrsel <= '0';
          if mute = '0' then
            flip_s <= not flip_s;
          end if;
        else
          block_count <= block_count + 1;
        end if;
        ata_reg <= "10111";  -- cmd_stat_reg
        fast_read <= '0';
        exec_state <= READ;
        next_state <= WAIT_BUSY_C_2; -- get response loop
      else
        if adrsel = '1' then
          bram_addr <= bram_addr+4; -- every 2 samples
        end if;
        adrsel <= not adrsel;
        size <= size - 1;
        delay_count <= "00111";
        exec_state <= READ;
        next_state <= GET_RESP;
      end if;
    ---------------------------------------------

    when DONE =>
      err_code_s <= "000";  -- ok
      exec_state <= IDLE;

    when ERR =>
      exec_state <= IDLE;  -- abort
     
    when READ =>
      dior_s <= '0';
      if delay_count = "00001" then
        data_in <= data;
        if mute = '0' then
          sample <= data;
        else
          sample <= (others => '0');
        end if;
        if fast_read = '1' then
          delay_count <= "00010";
        else
          delay_count <= "00000";
        end if;
        exec_state <= NEXT_STAT;
        if timeout_valid = '1' then
          if timeout = "00" & x"00000" then
            fast_read <= '0';
            err_code_s <= "100";
            exec_state <= ERR;
          end if;
          timeout <= timeout - 1;
        end if;
      else
        delay_count <= delay_count - 1;
      end if;
    
    when WRITE =>
      data_dir <= '0';  -- output
      diow_s <= '0';
      if delay_count = "00001" then
        exec_state <= NEXT_STAT;
      end if;
      delay_count <= delay_count - 1;
  
    when NEXT_STAT =>
      if delay_count = "00001" then
        exec_state <= next_state;
      end if;
      delay_count <= delay_count - 1;
   
    end case;

  end if;
end process;

end Behavioral;
